//  File:  PPGame.java
//
//  Demonstrate synchronization using wait and notifyAll

class PingPong {

    // State variable identifying whose turn it is
    private String whoseTurn = null;

    private static int nCount = 0;

    // return true to continue playing; false to stop playing
    public synchronized boolean hit(String opponent) {
    
        boolean bContinuePlaying = true;
        String x = Thread.currentThread().getName();
        
        if (whoseTurn == null) {
            whoseTurn = x;

        } else if (whoseTurn.compareTo("DONE") == 0)
            bContinuePlaying = false;

        else if (opponent.compareTo("DONE") == 0) {
            whoseTurn = opponent;
            notifyAll();
            bContinuePlaying = false;
        }

        else if (x.compareTo(whoseTurn) == 0) {
            System.out.println("PING! " + ++nCount + " (" + x + ")");
            whoseTurn = opponent;
            notifyAll();

        } else {
            try {
                long t1 = System.currentTimeMillis();
                wait(2500);
                if ((System.currentTimeMillis() - t1) > 2500) {
                    System.out.println("****** TIMEOUT! " + x 
                        + " is waiting to play");
                }
            } catch (InterruptedException e) { }
        }

        return bContinuePlaying;
    }

 }


class Player implements Runnable {

    PingPong myTable;        // Table where they play
    String myOpponent;
    
    public Player(String opponent, PingPong table) {
        myTable = table;
        myOpponent = opponent;
    }
        
    public void run() {
        while (myTable.hit(myOpponent))
            ;
    }

}


public class PPGame {

    public static void main(String[ ] args) {
        System.out.println("PingPong Game Begins");
        PingPong table = new PingPong();
        Thread alice   = new Thread(new Player("Bob", table));
        Thread bob     = new Thread(new Player("Alice", table));
        
        alice.setName("Alice");
        bob.setName("Bob");
        alice.start();        // Alice starts playing
        bob.start();        // Bob starts playing

        try {
            // Let the players play a bit...
            Thread.currentThread().sleep(100);
        } catch (InterruptedException e) { }
            
        table.hit("DONE"); // Tell the players to quit their threads
        System.out.println("Game Over!");

        try {
            Thread.sleep(1);
            System.out.println("Alice is alive = " + alice.isAlive()
                              + "    Bob is alive = " + bob.isAlive());
            alice.join();
            bob.join();
        } catch (InterruptedException e) { }
        System.out.println("Go Home!");
    }

}

/* ..... Output from this program

PingPong Game Begins
PING! 1 (Alice)
PING! 2 (Bob)
PING! 3 (Alice)
PING! 4 (Bob)
PING! 5 (Alice)
PING! 6 (Bob)
PING! 7 (Alice)
PING! 8 (Bob)
PING! 9 (Alice)
PING! 10 (Bob)
PING! 11 (Alice)
PING! 12 (Bob)
PING! 13 (Alice)
PING! 14 (Bob)
PING! 15 (Alice)
PING! 16 (Bob)
PING! 17 (Alice)
PING! 18 (Bob)
PING! 19 (Alice)
PING! 20 (Bob)
PING! 21 (Alice)
PING! 22 (Bob)
PING! 23 (Alice)
PING! 24 (Bob)
PING! 25 (Alice)
PING! 26 (Bob)
PING! 27 (Alice)
PING! 28 (Bob)
PING! 29 (Alice)
PING! 30 (Bob)
PING! 31 (Alice)
PING! 32 (Bob)
PING! 33 (Alice)
PING! 34 (Bob)
PING! 35 (Alice)
PING! 36 (Bob)
PING! 37 (Alice)
PING! 38 (Bob)
PING! 39 (Alice)
PING! 40 (Bob)
PING! 41 (Alice)
PING! 42 (Bob)
PING! 43 (Alice)
PING! 44 (Bob)
PING! 45 (Alice)
PING! 46 (Bob)
PING! 47 (Alice)
PING! 48 (Bob)
PING! 49 (Alice)
PING! 50 (Bob)
PING! 51 (Alice)
PING! 52 (Bob)
PING! 53 (Alice)
PING! 54 (Bob)
PING! 55 (Alice)
PING! 56 (Bob)
PING! 57 (Alice)
Game Over!
Alice is alive = true    Bob is alive = false
Go Home!

..... */
